//
//  DataStorage.swift
//  calculator-history
//
//  Created by Илья Лошкарёв on 03.03.2018.
//  Copyright © 2018 SFEDU. All rights reserved.
//
import Foundation
import CoreData

/// Протокол типа поддерживающего запрос по умолчанию
protocol Fetchable: NSFetchRequestResult {
    static func fetchAll() -> NSFetchRequest<Self>?
}

/// Хранилище данных в контексте CoreData
class DataStorage<T>: NSObject, NSFetchedResultsControllerDelegate
where T: NSManagedObject & Fetchable {
    
    // Объекты храняться в контексте
    let context: NSManagedObjectContext!
    
    // Получить объекты можно из результатов запроса
    let fetchResults: NSFetchedResultsController<T>
    
    // Транслируем внутренние изменения представителю
    weak var delegate: DataStorageDelegate?
    
    init(with context: NSManagedObjectContext) {
        self.context = context
        
        // создаём запрос
        guard let request = T.fetchAll() else {
            fatalError("Unable to request \(T.self)")
        }
        // сортировка результатов обязательна
        request.sortDescriptors = [NSSortDescriptor(key: "date", ascending: false)]
        
        // создаём контроллер результатов
        fetchResults = NSFetchedResultsController(
            fetchRequest: request,   // запрос
            managedObjectContext: context, // контекст выполнения запроса
            sectionNameKeyPath: nil, // имя поля для разбиения результатов на секции
            cacheName: nil) // имя файла для кэширования результатов
        
        super.init()  // для доступа к self
        
        // становимся представителем для обработки изменений в контексте
        fetchResults.delegate = self
        
        // получаем результаты запроса
        do {
            try fetchResults.performFetch()
        }
        catch { fatalError("Unable to fetch \(T.self)") }
    }
    
    subscript(index: Int) -> T {
        return fetchResults.object(at: IndexPath(row: index, section: 0))
    }
    
    var count: Int {
        return fetchResults.sections?[0].numberOfObjects ?? 0
    }
    
    func remove(at index: Int) {
        context.delete(self[index] as NSManagedObject)
    }
    
    func append(_ element: T) {
        context.insert(element as NSManagedObject)
    }
    
    func clear() {
        for element in fetchResults.fetchedObjects ?? [] {
            context.delete(element as NSManagedObject)
        }
    }
    
    //
    // MARK: NSFetchedResultsControllerDelegate
    //
    
    func controllerWillChangeContent(_ controller: NSFetchedResultsController<NSFetchRequestResult>) {
        delegate?.dataStorageWillChange(self) // сообщаем о начале изменений (для множественного удаления)
    }
    
    func controller(_ controller: NSFetchedResultsController<NSFetchRequestResult>, didChange anObject: Any, at indexPath: IndexPath?, for type: NSFetchedResultsChangeType, newIndexPath: IndexPath?) {
        // ретранслируем полученные изменения
        switch  type {
        case .delete:
            delegate?.dataStorage(self, didDeleteAt: indexPath!)
        case .insert:
            delegate?.dataStorage(self, didInsertAt: newIndexPath!)
        default:
            print("Unexpected change in context \(context.userInfo)")
        }
    }
    
    func controllerDidChangeContent(_ controller: NSFetchedResultsController<NSFetchRequestResult>) {
        delegate?.dataStorageDidChange(self) // сообщаем о завершении изменений (для множественного удаления)
    }
}
